@@ -1,7 +1,6 @@
 # coding: utf-8
 
-from __future__ import absolute_import
-from __future__ import print_function
+from __future__ import absolute_import, print_function
 
 """
 stuff to deal with comments and formatting on dict/list/ordereddict/set
@@ -9,17 +8,13 @@ these are not really related, formatting could be factored out as
 a separate base
 """
 
-from collections import MutableSet
+from collections import MutableSet, Sized, Set  # type: ignore
+
+from ruamel.yaml.compat import ordereddict, PY2
 
 __all__ = ["CommentedSeq", "CommentedMap", "CommentedOrderedMap",
            "CommentedSet", 'comment_attrib', 'merge_attrib']
 
-
-try:
-    from .compat import ordereddict
-except ImportError:
-    from ruamel.yaml.compat import ordereddict
-
 comment_attrib = '_yaml_comment'
 format_attrib = '_yaml_format'
 line_col_attrib = '_yaml_line_col'
@@ -29,7 +24,7 @@ tag_attrib = '_yaml_tag'
 
 
 class Comment(object):
-    # sys.getsize tested the Comment objects, __slots__ make them bigger
+    # sys.getsize tested the Comment objects, __slots__ makes them bigger
     # and adding self.end did not matter
     attrib = comment_attrib
 
@@ -326,6 +321,87 @@ class CommentedSeq(list, CommentedBase):
         return pre_comments
 
 
+class CommentedMapView(Sized):
+
+    __slots__ = '_mapping',
+
+    def __init__(self, mapping):
+        self._mapping = mapping
+
+    def __len__(self):
+        count = len(self._mapping)
+        done = []  # list of processed merge items, kept for masking
+        for merged in getattr(self._mapping, merge_attrib, []):
+            for x in merged[1]:
+                if self._mapping._unmerged_contains(x):
+                    continue
+                for y in done:
+                    if x in y:
+                        break
+                else:
+                    count += 1
+            done.append(merged[1])
+        return count
+
+    def __repr__(self):
+        return '{0.__class__.__name__}({0._mapping!r})'.format(self)
+
+
+class CommentedMapKeysView(CommentedMapView, Set):
+
+    __slots__ = ()
+
+    @classmethod
+    def _from_iterable(self, it):
+        return set(it)
+
+    def __contains__(self, key):
+        return key in self._mapping
+
+    def __iter__(self):
+        # yield from self._mapping  # not in py27, pypy
+        for x in self._mapping:
+            yield x
+
+
+class CommentedMapItemsView(CommentedMapView, Set):
+
+    __slots__ = ()
+
+    @classmethod
+    def _from_iterable(self, it):
+        return set(it)
+
+    def __contains__(self, item):
+        key, value = item
+        try:
+            v = self._mapping[key]
+        except KeyError:
+            return False
+        else:
+            return v == value
+
+    def __iter__(self):
+        for key in self._mapping._keys():
+            yield (key, self._mapping[key])
+
+
+class CommentedMapValuesView(CommentedMapView):
+
+    __slots__ = ()
+
+    def __contains__(self, value):
+        for key in self._mapping:
+            if value == self._mapping[key]:
+                return True
+        return False
+
+    def __iter__(self):
+        print('xxy values_iter')
+        for key in self._mapping:
+            yield self._mapping[key]
+
+
 class CommentedMap(ordereddict, CommentedBase):
     __slots__ = [Comment.attrib, ]
 
@@ -389,6 +465,14 @@ class CommentedMap(ordereddict, CommentedBase):
             for x in vals:
                 self[x] = vals[x]
 
+    def insert(self, pos, key, value, comment=None):
+        """insert key value into given position
+        attach comment if provided
+        """
+        ordereddict.insert(self, pos, key, value)
+        if comment is not None:
+            self.yaml_add_eol_comment(comment, key=key)
+
     def mlget(self, key, default=None, list_ok=False):
         """multi-level get that expects dicts within dicts"""
         if not isinstance(key, list):
@@ -422,12 +506,147 @@ class CommentedMap(ordereddict, CommentedBase):
                     return merged[1][key]
             raise
 
+    def _unmerged_contains(self, key):
+        if ordereddict.__contains__(self, key):
+            return True
+
+    def __contains__(self, key):
+        if ordereddict.__contains__(self, key):
+            return True
+        # this will only work once the mapping/dict is built to completion
+        for merged in getattr(self, merge_attrib, []):
+            if key in merged[1]:
+                return True
+        return False
+
     def get(self, key, default=None):
         try:
             return self.__getitem__(key)
         except:
             return default
 
+    def __repr__(self):
+        if not hasattr(self, merge_attrib):
+            return ordereddict.__repr__(self)
+        return 'ordereddict(' + repr(list(self._items())) + ')'
+
+    def non_merged_items(self):
+        for x in ordereddict.__iter__(self):
+            yield x, ordereddict.__getitem__(self, x)
+
+    def __delitem__(self, key):
+        found = True
+        for merged in getattr(self, merge_attrib, []):
+            try:
+                del merged[1][key]
+                found = True
+            except KeyError:
+                pass
+        try:
+            ordereddict.__delitem__(self, key)
+        except KeyError:
+            if not found:
+                raise
+
+    def __iter__(self):
+        for x in ordereddict.__iter__(self):
+            yield x
+        done = []  # list of processed merge items, kept for masking
+        for merged in getattr(self, merge_attrib, []):
+            for x in merged[1]:
+                if ordereddict.__contains__(self, x):
+                    continue
+                for y in done:
+                    if x in y:
+                        break
+                else:
+                    yield x
+            done.append(merged[1])
+
+    def _keys(self):
+        for x in ordereddict.__iter__(self):
+            yield x
+        done = []  # list of processed merge items, kept for masking
+        for merged in getattr(self, merge_attrib, []):
+            for x in merged[1]:
+                if ordereddict.__contains__(self, x):
+                    continue
+                for y in done:
+                    if x in y:
+                        break
+                else:
+                    yield x
+            done.append(merged[1])
+
+    if PY2:
+        def keys(self):
+            return list(self._keys())
+
+        def iterkeys(self):
+            return self._keys()
+
+        def viewkeys(self):
+            return CommentedMapKeysView(self)
+    else:
+        def keys(self):
+            return CommentedMapKeysView(self)
+
+    def _values(self):
+        for x in ordereddict.__iter__(self):
+            yield ordereddict.__getitem__(self, x)
+        done = []  # list of processed merge items, kept for masking
+        for merged in getattr(self, merge_attrib, []):
+            for x in merged[1]:
+                if ordereddict.__contains__(self, x):
+                    continue
+                for y in done:
+                    if x in y:
+                        break
+                else:
+                    yield ordereddict.__getitem__(merged[1], x)
+            done.append(merged[1])
+
+    if PY2:
+        def values(self):
+            return list(self._values())
+
+        def itervalues(self):
+            return self._values()
+
+        def viewvalues(self):
+            return CommentedMapValuesView(self)
+    else:
+        def values(self):
+            return CommentedMapValuesView(self)
+
+    def _items(self):
+        for x in ordereddict.__iter__(self):
+            yield x, ordereddict.__getitem__(self, x)
+        done = []  # list of processed merge items, kept for masking
+        for merged in getattr(self, merge_attrib, []):
+            for x in merged[1]:
+                if ordereddict.__contains__(self, x):
+                    continue
+                for y in done:
+                    if x in y:
+                        break
+                else:
+                    yield x, ordereddict.__getitem__(merged[1], x)
+            done.append(merged[1])
+
+    if PY2:
+        def items(self):
+            return list(self._items())
+
+        def iteritems(self):
+            return self._items()
+
+        def viewitems(self):
+            return CommentedMapItemsView(self)
+    else:
+        def items(self):
+            return CommentedMapItemsView(self)
+
     @property
     def merge(self):
         if not hasattr(self, merge_attrib):
